(************************************************)
(* VRg MIDI Music System 1.12                   *)
(* Debug MIDI device. This MIDI device writes   *)
(* verbosed MIDI events stream to stdout.       *)
(* All number values are hexadecimal.           *)
(* (C) 1998-1999 Virtual Research               *)
(* Unit version : 1.12                          *)
(* Last updated : 21.Nov.1999                   *)
(************************************************)

{$IFDEF _DYNAMIC}
 {$IFNDEF __TMT__}
  {$UNDEF _DYNAMIC}
 {$ENDIF}
{$ENDIF}

{$IFNDEF _DYNAMIC}
unit DbgMIDI;
{$ELSE}
program DbgMIDI;
{$ENDIF}

{$I DEF32.INC}
{$IFNDEF _32BIT }
 {$F+}
 {$C FIXED PRELOAD PERMANENT}
{$ENDIF}
{$I release.opt}

interface
uses VRMMSDef;

const VERSION='1.12';
      VersionInfo:longint=$00010102;
var MIDI:TMIDI;

implementation
uses PStrings {$IFDEF _DYNAMIC} ,DLL {$ENDIF};

function InitDevice(Port:word):boolean;
begin
 WriteLn(Output,'MIDI device init');
 InitDevice:=true;
end;

procedure ChangeMelodicBank(NewBank:pointer);
begin
 WriteLn(Output,'Melodic instruments bank changed to ',
     HexStr(longint(NewBank) shr 16),': ',
     HexStr(longint(NewBank)));
end;

procedure ChangeDrumsBank(NewBank:pointer);
begin
 WriteLn(Output,'Percussion instruments bank changed to ',
     HexStr(longint(NewBank) shr 16),': ',
     HexStr(longint(NewBank)));
end;

procedure NoteOn(Channel,Note,Velocity:byte);
begin
 Write(Output,'Channel ',RightStr(HexStr(Channel),2),': ',
    'Note ON ',RightStr(HexStr(Note),2),' ',
    RightStr(HexStr(Velocity),2));
 if Channel=9 then WriteLn(Output,' (',PercussionNames[Note],')')
  else WriteLn(Output);
end;

procedure NoteOff(Channel,Note,Velocity:byte);
begin
 Write(Output,'Channel ',RightStr(HexStr(Channel),2),': ',
    'Note OFF ',RightStr(HexStr(Note),2),' ',
    RightStr(HexStr(Velocity),2));
 if Channel=9 then WriteLn(Output,' (',PercussionNames[Note],')')
  else WriteLn(Output);
end;

procedure KeyAfterTouch(Channel,Note,Velocity:byte);
begin
 WriteLn(Output,'Channel ',RightStr(HexStr(Channel),2),': ',
    'Key aftertouch ',RightStr(HexStr(Note),2),' ',
    RightStr(HexStr(Velocity),2));
end;

procedure ChangeController(Channel,Controller,NewValue:byte);
var ControlName:string;
begin
 Write(Output,'Channel ',RightStr(HexStr(Channel),2),': ',
    'Change controller ',RightStr(HexStr(Controller),2));
 case Controller of
  MODULATION_POT:ControlName:='Modulation Pot';
  VOLUME_SET:ControlName:='Set Volume';
  PAN_SET:ControlName:='Set Panning';
  EXPRESSION:ControlName:='Expression';
  SUSTAIN_PEDAL:ControlName:='Sustain Pedal';
  SOFT_PEDAL:ControlName:='Soft Pedal';
  ALL_SOUNDS_OFF:ControlName:='All Sounds Off';
  ALL_NOTES_OFF:ControlName:='All Notes Off';
  RESET_ALL:ControlName:='Reset All Controllers';
 else ControlName:='';
 end;
 if ControlName<>'' then Write(' (',ControlName,') ')
  else Write(' ');
 WriteLn(RightStr(HexStr(NewValue),2));
end;

procedure PitchWheel(Channel:byte;Value:word);
begin
 WriteLn(Output,'Channel ',RightStr(HexStr(Channel),2),': ',
    'Pitch wheel ',HexStr(Value));
end;

procedure ChangePatch(Channel,Instrument:byte);
begin
 WriteLn(Output,'Channel ',RightStr(HexStr(Channel),2),': ',
    'Change patch to ',RightStr(HexStr(Instrument),2),
    ' (',MelodicNames[Instrument],')');
end;

procedure ChannelAfterTouch(Channel,Velocity:byte);
begin
 WriteLn(Output,'Channel ',RightStr(HexStr(Channel),2),': ',
    'Channel aftertouch ',RightStr(HexStr(Velocity),2));
end;

procedure ShutDownDevice;
begin
 WriteLn(Output,'MIDI device shut down');
end;

{$IFDEF _DYNAMIC}

const EXPORTED_OBJECTS=2;

      ExportTable:array[0..EXPORTED_OBJECTS] of TExportTableEntry=(
       (Name:'MIDI';ParmCount:0;Address:@MIDI),
       (Name:'VersionInfo';ParmCount:0;Address:@VersionInfo),
       (Name:nil;ParmCount:0;Address:nil));

      ExportTableHeader:TExportTableHeader=(
       Signature:'%EXPORT%';
       Table:@ExportTable;
       DummyString:nil);

{$ENDIF}

begin
 @MIDI.InitDevice:=@InitDevice;
 @MIDI.ChangeMelodicBank:=@ChangeMelodicBank;
 @MIDI.ChangeDrumsBank:=@ChangeDrumsBank;
 @MIDI.NoteOn:=@NoteOn;
 @MIDI.NoteOff:=@NoteOff;
 @MIDI.KeyAfterTouch:=@KeyAfterTouch;
 @MIDI.ChangeController:=@ChangeController;
 @MIDI.PitchWheel:=@PitchWheel;
 @MIDI.ChangePatch:=@ChangePatch;
 @MIDI.ChannelAfterTouch:=@ChannelAfterTouch;
 @MIDI.ShutDownDevice:=@ShutDownDevice;

{$IFDEF _DYNAMIC}
 ExportTableHeader.DummyString:=
  #13#10'Dynamic link driver for debug MIDI device (MIDI stream dumper)'+
  #13#10'Version '+VERSION+
  #13#10'Virtual Research independent group production, 1998-1999';
{$ENDIF}

end.

