(********************************)
(* VRg MIDI Music System 1.12   *)
(* MID MIDI sequence player     *)
(* (C) 1999 Virtual Research    *)
(* Unit version : 1.01          *)
(* Last updated : 21.Nov.1999   *)
(********************************)

{$IFDEF _DYNAMIC}
 {$IFNDEF __TMT__}
  {$UNDEF _DYNAMIC}
 {$ENDIF}
{$ENDIF}

unit MID;
{$I DEF32.INC}
{$IFNDEF _32BIT}
 {$C FIXED PRELOAD PERMANENT}
{$ENDIF}
{$I release.opt}

interface

var MidPlaying:boolean;
    MidError:integer;
const MidLoop:boolean=false;  (* It is not recommended to change these      *)
      MidVolume:byte=64;      (* constants "manually", there are respective *)
      MidTimerRate:word=364;  (* procedures.                                *)

{$IFNDEF _DYNAMIC}
procedure MidInitDevice(Device,Port:word);
{$ELSE}
procedure MidInitDevice(Device,Port:word;PathToDriver:string);
{$ENDIF}
procedure MidShutDownDevice;
procedure MidLoad(FileName:string;FilePos:longint;var Buffer:pointer);
procedure MidFree(Buffer:pointer);
procedure MidInitSequence(var Buffer:pointer);
procedure MidSetTimerRate(TPS:word);
procedure MidSetVolume(Vol:byte);       (* Vol is 0..64 *)
procedure MidPlay(Loop:boolean);
procedure MidStop;
procedure MidPoll;


implementation
uses Timer,VRMMSDef,MIDInit;

const MaxTracks= {$IFNDEF _32BIT} 5040; {$ELSE} 65535; {$ENDIF}

type TMidHeader=record
 Signature:array[0..3] of char;         (* identifier 'MThd' *)
 HeaderLength:longint;                  (* Always 6 *)
 FileType:word;                 (* 0 - single track, 1 - synch. multitrack,
                                   2 - asynch. mulitrack *)
 Tracks:word;                   (* Number of tracks *)
 Division:word;                 (* ticks per quarter *)
end;
     PMidHeader=^TMidHeader;

     TTrackHeader=record
 Signature:array[0..3] of char;         (* identifier 'MTrk' *)
 TrackLength:longint;                   (* Big-endian *)
end;
     PTrackHeader=^TTrackHeader;

     TTrackStub=record
 Track:PTrackHeader;
 TrackPosition:longint;
 TicksToWait:longint;
 LastEvent:byte;
end;
     PTrackStub=^TTrackStub;

     TTrackArray=array[1..MaxTracks] of TTrackStub;
     PTrackArray=^TTrackArray;

     TMidBufferStub=record
 Signature:array[0..5] of char;         (* identifier 'MTstub' or 'MTst**' *)
 Header:PMidHeader;
 TrackArray:PTrackArray;
end;
     PMidBufferStub=^TMidBufferStub;

const MidBuffer:PMidBufferStub=nil;
      uSecPerQuarter:longint=500000;
      MidiTicksPerTimerTick:longint=135;         (* x256 *)
      FramesPerSecond:byte=24;                   (* SMPTE frames *)

var MIDI:PMIDI;

{$IFNDEF _DYNAMIC}
procedure MidInitDevice(Device,Port:word);
begin
 MidError:=NO_ERROR;
 MIDI:=MidiInit(Device);
 if MIDI=nil then MidError:=DEVICE_UNAVAILABLE
  else if not MIDI^.InitDevice(Port) then MidError:=DEVICE_INIT_ERROR;
end;
{$ELSE}
procedure MidInitDevice(Device,Port:word;PathToDriver:string);
begin
 MidError:=NO_ERROR;
 MIDI:=MidiInit(Device,PathToDriver);
 if MIDI=nil then MidError:=DEVICE_UNAVAILABLE
  else if not MIDI^.InitDevice(Port) then MidError:=DEVICE_INIT_ERROR;
end;
{$ENDIF}

procedure MidShutDownDevice;
begin
 MidError:=NO_ERROR;
 MIDI^.ShutDownDevice;
{$IFDEF _DYNAMIC}
 MidiShutDown;
{$ENDIF}
end;

function GetData(TrackNum:word):byte;
type pbyte=^byte;
var DataPtr:pbyte;
begin
{$IFNDEF _32BIT}
 DataPtr:=Ptr(Seg(MidBuffer^.TrackArray^[TrackNum].Track^),
              Ofs(MidBuffer^.TrackArray^[TrackNum].Track^)+
              MidBuffer^.TrackArray^[TrackNum].TrackPosition);
{$ELSE}
 DataPtr:=pbyte(dword(MidBuffer^.TrackArray^[TrackNum].Track)+
                MidBuffer^.TrackArray^[TrackNum].TrackPosition);
{$ENDIF}
 Inc(MidBuffer^.TrackArray^[TrackNum].TrackPosition);
 GetData:=DataPtr^;
end;


procedure CalcTickRatio;
var MidiTicksPerSecond:longint;
    TicksPerFrame:word;
begin
 with MidBuffer^.Header^ do begin
  if (Division and $8000)<>0 then begin
   FramesPerSecond:=(Division and $7F00) shr 8;
   TicksPerFrame:=Division and $00FF;
   MidiTicksPerSecond:=FramesPerSecond*TicksPerFrame;
  end
  else MidiTicksPerSecond:=(Division*1000000) div uSecPerQuarter;
  MidiTicksPerTimerTick:=MidiTicksPerSecond*256 div MidTimerRate;
                                                 (* Scale is x256 *)
 end;
end;

function GetVarLen(TrackNum:word):longint;
var VarLen:longint;
    MidByte:byte;
begin
 VarLen:=0;
 repeat
  MidByte:=GetData(TrackNum);
  VarLen:=(VarLen shl 7)+(MidByte and $7F);
 until (MidByte and $80)=0;
 GetVarLen:=VarLen;
end;

procedure BSwap(var X:longint);assembler;
asm
{$IFNDEF _32BIT}
 les di,X
 mov ax,es:[di]
 mov bx,es:[di+2]
 xchg al,ah
 xchg bl,bh
 mov es:[di],bx
 mov es:[di+2],ax
{$ELSE}
 mov edi,X
 mov ax,[edi]
 mov bx,[edi+2]
 xchg al,ah
 xchg bl,bh
 mov [edi],bx
 mov [edi+2],ax
{$ENDIF}
end;


var TracksFinished:word;
    ActiveTrack:word;
    ActiveTicks:longint;

procedure MidPoll;
var i:word;

 procedure ProcessEvents(TrackNum:word);
 var Event,Channel:byte;
     EventData:array[1..2] of byte;
     RunningMode:boolean;
     i,Dummy:longint;

  procedure ProcessMetaEvent(EventNum:byte);
  var i,Dummy:longint;
      t1,t2,t3:longint;
  begin
   Dummy:=GetVarLen(TrackNum);  (* Event data length *)
   case EventNum of
    END_OF_TRACK:begin
     Inc(TracksFinished);
     MidBuffer^.TrackArray^[TrackNum].TicksToWait:=MaxLongInt;
    end;
    SET_TEMPO:begin
     t1:=GetData(TrackNum);     (* These are just dummy variables *)
     t2:=GetData(TrackNum);
     t3:=GetData(TrackNum);
     uSecPerQuarter:=t1*65536+t2*256+t3;
     CalcTickRatio;
    end;
    else for i:=1 to Dummy do GetData(TrackNum);   (* All other metaevents *)
   end;                                            (* are ignored          *)
  end;

 begin
  with MidBuffer^.TrackArray^[TrackNum] do begin
   if MidPlaying then begin
    while TicksToWait<=MidiTicksPerTimerTick do begin
     Event:=GetData(TrackNum);
     if Event=META_EVENT then ProcessMetaEvent(GetData(TrackNum)) else begin
      if (Event and $80)=0 then begin
       RunningMode:=true;
       EventData[1]:=Event;
       Event:=LastEvent;
      end else begin
       LastEvent:=Event;
       RunningMode:=false;
      end;
      Channel:=Event and $0F;
      Event:=Event and $F0;
      case Event of
       NOTE_OFF:begin
        if RunningMode then EventData[2]:=GetData(TrackNum)
         else begin
          EventData[1]:=GetData(TrackNum);
          EventData[2]:=GetData(TrackNum);
         end;
        EventData[2]:=((EventData[2]*MidVolume) shr 6);
        MIDI^.NoteOff(Channel,EventData[1],EventData[2]);
       end;
       NOTE_ON:begin
        if RunningMode then EventData[2]:=GetData(TrackNum)
         else begin
          EventData[1]:=GetData(TrackNum);
          EventData[2]:=GetData(TrackNum);
         end;
        EventData[2]:=((EventData[2]*MidVolume) shr 6);
        MIDI^.NoteOn(Channel,EventData[1],EventData[2]);
       end;
       KEY_AFTER_TOUCH:begin
        if RunningMode then EventData[2]:=GetData(TrackNum)
         else begin
          EventData[1]:=GetData(TrackNum);
          EventData[2]:=GetData(TrackNum);
         end;
        MIDI^.KeyAfterTouch(Channel,EventData[1],EventData[2]);
       end;
       CONTROL_CHANGE:begin
        if RunningMode then EventData[2]:=GetData(TrackNum)
         else begin
          EventData[1]:=GetData(TrackNum);
          EventData[2]:=GetData(TrackNum);
         end;
        if EventData[1]=VOLUME_SET then
         EventData[2]:=((EventData[2]*MidVolume) shr 6);
        MIDI^.ChangeController(Channel,EventData[1],EventData[2]);
       end;
       PATCH_CHANGE:begin
        if not RunningMode then EventData[1]:=GetData(TrackNum);
        MIDI^.ChangePatch(Channel,EventData[1]);
       end;
       CHANNEL_AFTER_TOUCH:begin
        if not RunningMode then EventData[1]:=GetData(TrackNum);
        MIDI^.ChannelAfterTouch(Channel,EventData[1]);
       end;
       PITCH_WHEEL:begin
        if RunningMode then EventData[2]:=GetData(TrackNum)
         else begin
          EventData[1]:=GetData(TrackNum);
          EventData[2]:=GetData(TrackNum);
         end;
        MIDI^.PitchWheel(Channel,EventData[1]+EventData[2]*256);
       end;
       SYS_EX:begin
        Dummy:=GetVarLen(TrackNum);
        for i:=1 to Dummy do GetData(TrackNum);
       end;
      end;
     end;
     if TicksToWait<>MaxLongInt then   (* Finished track has MaxLongInt
                                          ticks to wait *)
               (* Get delay to next event *)
      TicksToWait:=GetVarLen(TrackNum)*256 else Break;
    end;
   end;
  end;
 end;

begin
 if MidPlaying then begin
 if TracksFinished>=MidBuffer^.Header^.Tracks then begin
         (* All tracks finished *)
  if MidLoop then begin
   for i:=1 to MidBuffer^.Header^.Tracks do begin
    MidBuffer^.TrackArray^[i].TrackPosition:=SizeOf(TTrackHeader);
    MidBuffer^.TrackArray^[i].TicksToWait:=GetVarLen(i)*256;
    MidBuffer^.TrackArray^[i].LastEvent:=0;
   end;
   TracksFinished:=0;
   ActiveTrack:=0;
   ActiveTicks:=MaxLongInt;
  end
  else MidPlaying:=false;
 end
 else begin
  case MidBuffer^.Header^.FileType of
   0: begin
    with MidBuffer^.TrackArray^[1] do
     if (TicksToWait>MidiTicksPerTimerTick) and
        (TicksToWait<>MaxLongInt)  then
      Dec(TicksToWait,MidiTicksPerTimerTick) else ProcessEvents(1);
   end;
   1: begin
           (* The following mess is a very tricky method
              used to synchronize tracks. If you don't understand
              how it works, don't worry, - neither do I :) *)
    if ActiveTrack=0 then begin
     for i:=1 to MidBuffer^.Header^.Tracks do begin (* Find active track *)
      if (MidBuffer^.TrackArray^[i].TicksToWait<ActiveTicks) and
         (MidBuffer^.TrackArray^[i].TicksToWait<>MaxLongInt) then begin
       ActiveTrack:=i;
       ActiveTicks:=MidBuffer^.TrackArray^[i].TicksToWait;
      end;
     end;
    end;
    with MidBuffer^.TrackArray^[ActiveTrack] do begin
     if (TicksToWait>MidiTicksPerTimerTick) and
        (TicksToWait<>MaxLongInt) then
      Dec(TicksToWait,MidiTicksPerTimerTick) else begin
      TicksToWait:=ActiveTicks;      (* Just in case.. *)
      for i:=1 to MidBuffer^.Header^.Tracks do begin
       Dec(MidBuffer^.TrackArray^[i].TicksToWait,ActiveTicks);
       ProcessEvents(i);
      end;
      ActiveTrack:=0; ActiveTicks:=MaxLongInt;
     end;
    end;
   end;
   2: begin
    with MidBuffer^.TrackArray^[TracksFinished+1] do
     if (TicksToWait>MidiTicksPerTimerTick) and
        (TicksToWait<>MaxLongInt)  then
      Dec(TicksToWait,MidiTicksPerTimerTick) else
       ProcessEvents(TracksFinished+1);
   end;
  end;
 end;
end;
end;

procedure MidPlay(Loop:boolean);
begin
 MidPlaying:=true;
 MidLoop:=Loop;
 TimerInstallHandler(@MidPoll);
 TimerSetup(MidTimerRate);
end;

procedure MidStop;
begin
 MidPlaying:=false;
 TimerDone;
 TimerRemoveHandler;
end;

procedure MidSetTimerRate(TPS:word);
begin
 MidTimerRate:=TPS;
 if MidPlaying then begin
  MidStop;
  CalcTickRatio;
  MidPlay(MidLoop);
 end else CalcTickRatio;
end;

procedure MidSetVolume(Vol:byte);
var i:byte;
begin
 if Vol>64 then Vol:=64;
 MidVolume:=Vol;
end;

procedure MidLoad(FileName:string;FilePos:longint;var Buffer:pointer);
var f:file;
    Dummy: {$IFDEF _32BIT} longint; {$ELSE} word; {$ENDIF}
    TrackHeader:TTrackHeader;
    i:word;

 procedure CleanUpStub;
 begin
  Close(f);
  FreeMem(PMidBufferStub(Buffer)^.Header,SizeOf(TMidHeader));
  FreeMem(Buffer,SizeOf(TMidBufferStub));
 end;

 procedure CleanUpTracks;
 var j:word;
 begin
  with PMidBufferStub(Buffer)^ do begin
   for j:=1 to i-1 do FreeMem(TrackArray^[i].Track,
         TrackArray^[i].Track^.TrackLength+SizeOf(TTrackHeader));
   FreeMem(TrackArray,Header^.Tracks*SizeOf(TTrackStub));
  end;
 end;

begin
 MidError:=NO_ERROR;
 Assign(f,FileName);
{$I-}
 Dummy:=FileMode;
 Filemode:=0;
 Reset(f,1);
 FileMode:=Dummy;
 if IOResult<>0 then begin
  MidError:=LOAD_ERROR;
  Exit;
 end;
 Seek(f,FilePos);
 if IOResult<>0 then begin
  Close(f);
  MidError:=LOAD_ERROR;
  Exit;
 end;
{$I+}
 if MaxAvail<(SizeOf(TMidBufferStub)+SizeOf(TMidHeader)) then begin
  Close(f);
  MidError:=NOT_ENOUGH_MEMORY;
  Exit;
 end;
 GetMem(Buffer,SizeOf(TMidBufferStub));
 with PMidBufferStub(Buffer)^ do begin
  Signature:='MTstub';
  GetMem(Header,SizeOf(TMidHeader));
  BlockRead(f,Header^,SizeOf(TMidHeader),Dummy);
  with Header^ do begin
   if Signature<>'MThd' then begin
    CleanUpStub;
    MidError:=INVALID_SEQ_FORMAT;
    Exit;
   end;
   Bswap(HeaderLength);
   FileType:=Swap(FileType);
   Tracks:=Swap(Tracks);
   Division:=Swap(Division);
{$IFNDEF _32BIT}
   if Tracks>MaxTracks then begin
    CleanUpStub;
    MidError:=SEQ_TOO_LARGE;
    Exit;
   end;
{$ENDIF}
   if MaxAvail<Tracks*SizeOf(TTrackStub) then begin
    CleanUpStub;
    MidError:=NOT_ENOUGH_MEMORY;
    Exit;
   end;
   GetMem(TrackArray,Tracks*SizeOf(TTrackStub));
  end;
  Seek(f,FilePos+SizeOf(TMidHeader)+Header^.HeaderLength-6);
  for i:=1 to Header^.Tracks do begin
   BlockRead(f,TrackHeader,SizeOf(TTrackHeader),Dummy);
   if TrackHeader.Signature<>'MTrk' then begin
    CleanUpTracks;
    CleanUpStub;
    MidError:=INVALID_SEQ_FORMAT;
    Exit;
   end;
   Bswap(TrackHeader.TrackLength);
   if MaxAvail<TrackHeader.TrackLength then begin
    CleanUpTracks;
    CleanUpStub;
    MidError:=NOT_ENOUGH_MEMORY;
    Exit;
   end;
{$IFNDEF _32BIT}
   if TrackHeader.TrackLength>65520 then begin
    CleanUpTracks;
    CleanUpStub;
    MidError:=SEQ_TOO_LARGE;
    Exit;
   end;
{$ENDIF}
   GetMem(TrackArray^[i].Track,
          TrackHeader.TrackLength+SizeOf(TTrackHeader));
   Seek(f,System.FilePos(f)-SizeOf(TTrackHeader));
   BlockRead(f,TrackArray^[i].Track^,
             TrackHeader.TrackLength+SizeOf(TTrackHeader),Dummy);
   Bswap(TrackArray^[i].Track^.TrackLength);
   TrackArray^[i].TrackPosition:=SizeOf(TTrackHeader);
   TrackArray^[i].LastEvent:=0;
  end;
 end;
 Close(f);
end;

procedure MidInitSequence(var Buffer:pointer);
var Stub:PMidBufferStub;
    TrackPtr:PTrackHeader;
    i:word;
begin
 MidError:=NO_ERROR;
 if PMidBufferStub(Buffer)^.Signature<>'MTstub' then begin
  if PMidHeader(Buffer)^.Signature<>'MThd' then begin
   MidError:=INVALID_SEQ_FORMAT;
   Exit;
  end;
  GetMem(Stub,SizeOf(TMidBufferStub));
  with Stub^ do begin
   Signature:='MTst**';
   GetMem(Header,SizeOf(TMidHeader));
   Header^:=TMidHeader(Buffer^);
   with Header^ do begin
    Bswap(HeaderLength);
    FileType:=Swap(FileType);
    Tracks:=Swap(Tracks);
    Division:=Swap(Division);
{$IFNDEF _32BIT}
    if Tracks>MaxTracks then begin
     FreeMem(Header,SizeOf(TMidHeader));
     FreeMem(Stub,SizeOf(TMidBufferStub));
     MidError:=SEQ_TOO_LARGE;
     Exit;
    end;
{$ENDIF}
    if MaxAvail<Tracks*SizeOf(TTrackStub) then begin
     FreeMem(Header,SizeOf(TMidHeader));
     FreeMem(Stub,SizeOf(TMidBufferStub));
     MidError:=NOT_ENOUGH_MEMORY;
     Exit;
    end;
    GetMem(TrackArray,Tracks*SizeOf(TTrackStub));
   end;
{$IFNDEF _32BIT}
   TrackPtr:=Ptr(Seg(Buffer^),
             Ofs(Buffer^)+SizeOf(TMidHeader)+Header^.HeaderLength-6);
{$ELSE}
   TrackPtr:=PTrackHeader(Buffer+SizeOf(TMidHeader)+Header^.HeaderLength-6);
{$ENDIF}
   for i:=1 to Header^.Tracks do begin
    if TrackPtr^.Signature<>'MTrk' then begin
     FreeMem(TrackArray,Header^.Tracks*SizeOf(TTrackStub));
     FreeMem(Header,SizeOf(TMidHeader));
     FreeMem(Stub,SizeOf(TMidBufferStub));
     MidError:=INVALID_SEQ_FORMAT;
     Exit;
    end;
    Bswap(TrackPtr^.TrackLength);
    TrackArray^[i].Track:=TrackPtr;
    TrackArray^[i].TrackPosition:=SizeOf(TTrackHeader);
    TrackArray^[i].LastEvent:=0;
    Inc(TrackPtr,TrackPtr^.TrackLength);
   end;
  end;
  Buffer:=Stub;
 end;
 MidBuffer:=Buffer;
 for i:=1 to MidBuffer^.Header^.Tracks do
  MidBuffer^.TrackArray^[i].TicksToWait:=GetVarLen(i)*256;
 if MidBuffer^.Header^.FileType>2 then MidBuffer^.Header^.FileType:=2;
 CalcTickRatio;
 TracksFinished:=0;
 ActiveTrack:=0;
 ActiveTicks:=MaxLongInt;
end;

procedure MidFree(Buffer:pointer);
var i:word;
begin
 MidError:=NO_ERROR;
 with PMidBufferStub(Buffer)^ do begin
  if Signature='MTstub' then begin
   for i:=1 to Header^.Tracks do FreeMem(TrackArray^[i].Track,
         TrackArray^[i].Track^.TrackLength+SizeOf(TTrackHeader));
  end;
  if (Signature='MTstub') or (Signature='MTst**') then begin
   FreeMem(TrackArray,Header^.Tracks*SizeOf(TTrackStub));
   FreeMem(Header,SizeOf(TMidHeader));
   FreeMem(Buffer,SizeOf(TMidBufferStub));
  end;
 end;
end;

begin
 MidPlaying:=false;
 MidError:=NO_ERROR;
end.

