(************************************************)
(* General MIDI (MPU401) interface              *)
(* for Borland Pascal 7.0 / TMT Pascal compiler *)
(* (C) 1998 by Asp                              *)
(* Virtual Research i. g. production            *)
(*                                              *)
(* Unit version : 1.03                          *)
(* Started      : 2.Jun.1998                    *)
(* Finished     : 3.Jun.1998                    *)
(* Last updated : 21.Nov.1999                   *)
(* System       : DOS/DPMI16/DPMI32             *)
(************************************************)

unit MPU401;
{$I DEF32.INC}
{$I release.opt}

interface

 const MPU401_IO_Addr:word=$330;        (* Must be set to actual MPU addr *)

 const gmError:array[1..3] of pchar=(
  'MPU401 not detected',
  'MPU401 device timeout',
  'Failed to enter UART mode');

 function  gmResetMPU:integer;
 procedure gmWriteMIDIData(data:byte);
 function  gmReadMIDIData:byte;
 procedure gmOutMIDICommand(command,channel:byte);


{$I GMEVENTS.INC}

implementation

 var OldExit:pointer;

 function ResetMPU:integer;assembler;   (* This may be used for detection *)
 asm
  mov dx,MPU401_IO_Addr
  inc dx
  xor bx,bx
@@busy:
  dec bx
  jz @@TimeOut
  in al,dx
  test al,40h
  jnz @@busy
  mov al,$FF
  out dx,al
  xor bx,bx
@@empty:
  dec bx
  jz @@TimeOut
  in al,dx
  test al,80h
  jnz @@empty

  dec dx
  in al,dx
  cmp al,$FE
  je @@Ok
  inc dx
  jmp @@NotFound

@@TimeOut:
  mov ax,2
  jmp @@exit

@@NotFound:
  mov ax,1
  jmp @@exit

@@Ok:
  xor ax,ax
@@exit:
 end;

 function EnterUART:integer;assembler;      (* Sets MPU401 to UART 
                                               (asynchronous port) mode.
                                               In SB AWE32/64 this is the
                                               only available mode. *)
 asm
  mov dx,MPU401_IO_Addr
  inc dx
  xor bx,bx
@@busy:
  dec bx
  jz @@TimeOut
  in al,dx
  test al,40h
  jnz @@busy
  mov al,$3F
  out dx,al
  xor bx,bx
@@empty:
  dec bx
  jz @@TimeOut
  in al,dx
  test al,80h
  jnz @@empty

  dec dx
  in al,dx
  cmp al,$FE
  je @@Ok
  inc dx
  jmp @@Failed

@@TimeOut:
  mov ax,2
  jmp @@exit

@@Failed:
  mov ax,3
  jmp @@exit

@@Ok:
  xor ax,ax
@@exit:
 end;

 function gmResetMPU:integer;
 var Retcode:integer;
 begin
  Retcode:=ResetMPU;
  if Retcode<>0 then gmResetMPU:=Retcode else gmResetMPU:=EnterUART;
 end;

 procedure gmWriteMIDIData(data:byte);assembler;
 asm
  mov dx,MPU401_IO_Addr
  inc dx
  xor bx,bx
@@busy:
  dec bx
  jz @@TimeOut
  in al,dx
  test al,40h
  jnz @@busy
@@TimeOut:
  dec dx
  mov al,data
  out dx,al
 end;

 function gmReadMIDIData:byte;assembler;
 asm
  mov dx,MPU401_IO_Addr
  inc dx
  xor bx,bx
@@busy:
  dec bx
  jz @@TimeOut
  in al,dx
  test al,80h
  jnz @@busy
@@TimeOut:
  dec dx
  in al,dx
 end;

 procedure gmOutMIDICommand(command,channel:byte);
 begin
  gmWriteMIDIData(command+channel);
 end;

 procedure gmExitProc; {$IFNDEF _32BIT} far; {$ENDIF}
 var i:byte;
 begin
  ExitProc:=OldExit;
  for i:=0 to 15 do begin
   gmOutMIDICommand(CONTROL_CHANGE,i);
   gmWriteMIDIData($78); (* All sounds off *)
   gmWriteMIDIData(0);
  end;
  ResetMPU;
 end;

begin
 OldExit:=ExitProc;
 ExitProc:=@gmExitProc;
end.

