(********************************************)
(* VRg MIDI Music System 1.12               *)
(* MPU401-based general MIDI device driver  *)
(* (C) 1998-1999 Virtual Research           *)
(* Unit version : 1.12                      *)
(* Last updated : 21.Nov.1999               *)
(********************************************)

{$IFDEF _DYNAMIC}
 {$IFNDEF __TMT__}
  {$UNDEF _DYNAMIC}
 {$ENDIF}
{$ENDIF}

{$IFNDEF _DYNAMIC}
unit MPUMIDI;
{$ELSE}
program MPUMIDI;
{$ENDIF}

{$I DEF32.INC}
{$IFNDEF _32BIT}
 {$F+}
 {$C FIXED PRELOAD PERMANENT}
{$ENDIF}
{$I release.opt}

interface
uses VRMMSDef;

const VERSION='1.12';
      VersionInfo:longint=$00010102;

var MIDI:TMIDI;

implementation
uses MPU401 {$IFDEF _DYNAMIC} ,DLL {$ENDIF};

function InitDevice(Port:word):boolean;
begin
 MPU401_IO_Addr:=Port;
 if gmResetMPU<>0 then InitDevice:=false
 else InitDevice:=true;
end;

procedure ChangeMelodicBank(Bank:pointer);
begin
end;

procedure ChangeDrumsBank(Bank:pointer);
begin
end;

procedure NoteOn(Channel,Note,Velocity:byte);
begin
 gmOutMIDICommand(NOTE_ON,Channel);
 gmWriteMIDIData(Note);
 gmWriteMIDIData(Velocity);
end;

procedure NoteOff(Channel,Note,Velocity:byte);
begin
 gmOutMIDICommand(NOTE_OFF,Channel);
 gmWriteMIDIData(Note);
 gmWriteMIDIData(Velocity);
end;

procedure KeyAfterTouch(Channel,Note,Velocity:byte);
begin
 gmOutMIDICommand(KEY_AFTER_TOUCH,Channel);
 gmWriteMIDIData(Note);
 gmWriteMIDIData(Velocity);
end;

procedure ChangeController(Channel,Controller,NewValue:byte);
begin
 gmOutMIDICommand(CONTROL_CHANGE,Channel);
 gmWriteMIDIData(Controller);
 gmWriteMIDIData(NewValue);
end;

procedure PitchWheel(Channel:byte;Value:word);
begin
 gmOutMIDICommand(PITCH_WHEEL,Channel);
 gmWriteMIDIData(Hi(Value));
 gmWriteMIDIData(Lo(Value));
end;

procedure ChangePatch(Channel,Instrument:byte);
begin
 gmOutMIDICommand(PATCH_CHANGE,Channel);
 gmWriteMIDIData(Instrument);
end;

procedure ChannelAfterTouch(Channel,Velocity:byte);
begin
 gmOutMIDICommand(CHANNEL_AFTER_TOUCH,Channel);
 gmWriteMIDIData(Velocity);
end;

procedure ShutDownDevice;
var i:byte;
begin
 for i:=0 to 15 do begin
  gmOutMIDICommand(CONTROL_CHANGE,i);
  gmWriteMIDIData($78); (* All sounds off *)
  gmWriteMIDIData(0);
 end;
end;

{$IFDEF _DYNAMIC}

const EXPORTED_OBJECTS=2;

      ExportTable:array[0..EXPORTED_OBJECTS] of TExportTableEntry=(
       (Name:'MIDI';ParmCount:0;Address:@MIDI),
       (Name:'VersionInfo';ParmCount:0;Address:@VersionInfo),
       (Name:nil;ParmCount:0;Address:nil));

      ExportTableHeader:TExportTableHeader=(
       Signature:'%EXPORT%';
       Table:@ExportTable;
       DummyString:nil);

{$ENDIF}

begin
 @MIDI.InitDevice:=@InitDevice;
 @MIDI.ChangeMelodicBank:=@ChangeMelodicBank;
 @MIDI.ChangeDrumsBank:=@ChangeDrumsBank;
 @MIDI.NoteOn:=@NoteOn;
 @MIDI.NoteOff:=@NoteOff;
 @MIDI.KeyAfterTouch:=@KeyAfterTouch;
 @MIDI.ChangeController:=@ChangeController;
 @MIDI.PitchWheel:=@PitchWheel;
 @MIDI.ChangePatch:=@ChangePatch;
 @MIDI.ChannelAfterTouch:=@ChannelAfterTouch;
 @MIDI.ShutDownDevice:=@ShutDownDevice;

{$IFDEF _DYNAMIC}
 ExportTableHeader.DummyString:=
  #13#10'Dynamic link driver for MPU401 UART MIDI port'+
  #13#10'Version '+VERSION+
  #13#10'Virtual Research independent group production, 1998-1999';
{$ENDIF}

end.
