(************************************************************)
(* VRg MIDI Music System 1.12                               *)
(* MUS MIDI sequence player.                                *)
(*      Surely not the best music player you've heard :)    *)
(* (C) 1998-1999 Virtual Research                           *)
(* Unit version : 1.12                                      *)
(* Last updated : 21.Nov.1999                               *)
(************************************************************)

{$IFDEF _DYNAMIC}
 {$IFNDEF __TMT__}
  {$UNDEF _DYNAMIC}
 {$ENDIF}
{$ENDIF}

unit MUS;
{$I DEF32.INC}
{$IFNDEF _32BIT}
 {$C FIXED PRELOAD PERMANENT}
{$ENDIF}
{$I release.opt}

interface

var MusPlaying:boolean;
    MusError:integer;
const MusLoop:boolean=false;  (* It is not recommended to change these      *)
      MusVolume:byte=64;      (* constants "manually", there are respective *)
      MusTempo:word=140;      (* procedures.                                *)
      MusTimerRate:word=182;

{$IFNDEF _DYNAMIC}
procedure MusInitDevice(Device,Port:word);
{$ELSE}
procedure MusInitDevice(Device,Port:word;PathToDriver:string);
{$ENDIF}
procedure MusShutDownDevice;
procedure MusLoad(FileName:string;FilePos:longint;var Buffer:pointer;
                  var AbsoluteLength:word);
procedure MusInitSequence(Buffer:pointer);
procedure MusSetTempo(TPS:word);
procedure MusSetTimerRate(TPS:word);
procedure MusSetVolume(Vol:byte);       (* Vol is 0..64 *)
procedure MusPlay(Loop:boolean);
procedure MusStop;
procedure MusPoll;


implementation
uses Timer,VRMMSDef,MIDInit;

type TChannelInfo=record
 PhysicalChannel:byte;
 Volume:byte;
end;

const ChannelMap:array[0..15] of TChannelInfo=    (* Logical to physical
                                                   channels mapping *)
                  ((PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:0;Volume:64),
                   (PhysicalChannel:9;Volume:64));

      ControlMap:array[1..14] of byte=          (* MUS controller numbers
                                                   to MIDI controller
                                                   numbers mapping *)
      (0,32,7,10,11,91,93,64,67,120,123,126,127,121);

      (* MUS events *)
      NOTE_ON=1;
      NOTE_OFF=0;
      PITCH_WHEEL=2;
      SYSTEM_EVENT=3;
      CHANGE_CONTROLLER=4;
      SCORE_END=6;

      MusPosition:word=0;
      StartPosition:word=0;
      MusBuffer:pointer=nil;
      MusTicksPerTimerTick:longint=197;         (* x256 *)


type TMusHeader=record
 Signature:array[0..3] of char;         (* identifier 'MUS'#$1A *)
 ScoreLength:word;
 ScoreStart:word;
 Channels:word;
 SecondaryChannels:word;
 InstrCnt:word;
 InstrList:word;
end;
     PMusHeader=^TMusHeader;

var MIDI:PMIDI;

{$IFNDEF _DYNAMIC}
procedure MusInitDevice(Device,Port:word);
begin
 MusError:=NO_ERROR;
 MIDI:=MidiInit(Device);
 if MIDI=nil then MusError:=DEVICE_UNAVAILABLE
  else if not MIDI^.InitDevice(Port) then musError:=DEVICE_INIT_ERROR;
end;
{$ELSE}
procedure MusInitDevice(Device,Port:word;PathToDriver:string);
begin
 MusError:=NO_ERROR;
 MIDI:=MidiInit(Device,PathToDriver);
 if MIDI=nil then MusError:=DEVICE_UNAVAILABLE
  else if not MIDI^.InitDevice(Port) then musError:=DEVICE_INIT_ERROR;
end;
{$ENDIF}

procedure MusShutDownDevice;
begin
 MusError:=NO_ERROR;
 MIDI^.ShutDownDevice;
{$IFDEF _DYNAMIC}
 MidiShutDown;
{$ENDIF}
end;

function GetData:byte;assembler;
{$IFNDEF _32BIT}
asm
 les di,MusBuffer
 add di,MusPosition
 mov al,es:[di]
 inc MusPosition
end;
{$ELSE}
asm
 movzx eax,MusPosition
 mov edi,MusBuffer
 add edi,eax
 mov al,[edi]
 inc MusPosition
end;
{$ENDIF}

function GetDelay:longint;
var Time:longint;
    MusByte:byte;
begin
 Time:=0;
 repeat
  MusByte:=GetData;
  Time:=(Time shl 7)+(MusByte and $7F);
 until (MusByte and $80)=0;
 GetDelay:=Time;
end;


var TicksToWait:longint;

procedure MusPoll;              (* This is the actual player *)
var Event,Channel:byte;
    EventData:array[1..2] of byte;
    GetDelayFlag:boolean;
begin
 if TicksToWait>MusTicksPerTimerTick then begin
  Dec(TicksToWait,MusTicksPerTimerTick);
  Exit;
 end
 else TicksToWait:=0;
 if MusPlaying then begin
  while TicksToWait=0 do begin
   Event:=GetData;
   if (Event and $80)<>0 then GetDelayFlag:=true else GetDelayFlag:=false;
   Channel:=Event and $0F;
   Event:=(Event and $70) shr 4;
   case Event of
    NOTE_ON: begin
     EventData[1]:=GetData;
     if (EventData[1] and $80)<>0 then begin
      EventData[1]:=EventData[1] and $7F;
      EventData[2]:=GetData;
      ChannelMap[Channel].Volume:=EventData[2];
      EventData[2]:=(MusVolume*EventData[2]) shr 6;
     end
     else EventData[2]:=(MusVolume*ChannelMap[Channel].Volume) shr 6;
     MIDI^.NoteOn(ChannelMap[Channel].PhysicalChannel,
                  EventData[1],EventData[2]);
    end;
    NOTE_OFF: begin
     EventData[1]:=GetData;
     MIDI^.NoteOff(ChannelMap[Channel].PhysicalChannel,
                   EventData[1],(MusVolume*ChannelMap[Channel].Volume) shr 6);
    end;
    PITCH_WHEEL: begin
     EventData[1]:=GetData;
     MIDI^.PitchWheel(ChannelMap[Channel].PhysicalChannel,
                      64*256+(EventData[1] shr 1));
    end;
    SYSTEM_EVENT: begin
     EventData[1]:=GetData;
     MIDI^.ChangeController(ChannelMap[Channel].PhysicalChannel,
                            ControlMap[EventData[1]],0);
    end;
    CHANGE_CONTROLLER: begin
     EventData[1]:=GetData;
     EventData[2]:=GetData;
     if EventData[1]=3 then EventData[2]:=(MusVolume*EventData[2]) shr 6;
     if EventData[1]=0 then MIDI^.ChangePatch(ChannelMap[Channel].PhysicalChannel,
                                              EventData[2]) else
      MIDI^.ChangeController(ChannelMap[Channel].PhysicalChannel,
                             ControlMap[EventData[1]],EventData[2]);
    end;
    SCORE_END: begin
     for Channel:=0 to 15 do ChannelMap[Channel].Volume:=MusVolume;
     if not MusLoop then MusPlaying:=false
     else MusPosition:=StartPosition;
    end;
   end;
   if GetDelayFlag then TicksToWait:=GetDelay*256;
  end;
 end;
end;

procedure MusPlay(Loop:boolean);
begin
 MusPlaying:=true;
 MusLoop:=Loop;
 TimerInstallHandler(@MusPoll);
 TimerSetup(MusTimerRate);
end;

procedure MusStop;
begin
 MusPlaying:=false;
 TimerDone;
 TimerRemoveHandler;
end;

procedure MusSetTempo(TPS:word);
begin
 MusTempo:=TPS;
 if MusPlaying then begin
  MusStop;
  MusTicksPerTimerTick:=(MusTempo shl 8) div MusTimerRate;
  MusPlay(MusLoop);
 end
 else MusTicksPerTimerTick:=(MusTempo shl 8) div MusTimerRate;
end;

procedure MusSetTimerRate(TPS:word);
begin
 MusTimerRate:=TPS;
 if MusPlaying then begin
  MusStop;
  MusTicksPerTimerTick:=(MusTempo shl 8) div MusTimerRate;
  MusPlay(MusLoop);
 end
 else MusTicksPerTimerTick:=(MusTempo shl 8) div MusTimerRate;
end;

procedure MusSetVolume(Vol:byte);
var i:byte;
begin
 if Vol>64 then Vol:=64;
 MusVolume:=Vol;
end;

procedure MusLoad(FileName:string;FilePos:longint;var Buffer:pointer;
                  var AbsoluteLength:word);
var f:file;
    Header:TMusHeader;
    Dummy:word;
begin
 MusError:=NO_ERROR;
 Assign(f,FileName);
{$I-}
 Dummy:=FileMode;
 FileMode:=0;
 Reset(f,1);
 FileMode:=Dummy;
 if IOResult<>0 then begin
  MusError:=LOAD_ERROR;
  Exit;
 end;
 Seek(f,FilePos);
 if IOResult<>0 then begin
  Close(f);
  MusError:=LOAD_ERROR;
  Exit;
 end;
{$I+}
 BlockRead(f,Header,SizeOf(Header),Dummy);
 if Header.Signature<>'MUS'#$1A then begin
  Close(f);
  MusError:=INVALID_SEQ_FORMAT;
  Exit;
 end;
 AbsoluteLength:=Header.ScoreStart+Header.ScoreLength;
 if MaxAvail<AbsoluteLength then begin
  Close(f);
  MusError:=NOT_ENOUGH_MEMORY;
  Exit;
 end;
{$IFNDEF _32BIT}
 if AbsoluteLength>65520 then begin
  Close(f);
  MusError:=SEQ_TOO_LARGE;
  Exit;
 end;
{$ENDIF}
 GetMem(Buffer,AbsoluteLength);
 Seek(f,FilePos);
 BlockRead(f,Buffer^,AbsoluteLength,Dummy);
 Close(f);
end;

procedure MusInitSequence(Buffer:pointer);
var i:word;
begin
 MusError:=NO_ERROR;
 if PMusHeader(Buffer)^.Signature='MUS'#$1A then begin
  if PMusHeader(Buffer)^.Channels<10 then
   for i:=0 to PMusHeader(Buffer)^.Channels-1 do
    ChannelMap[i].PhysicalChannel:=i
  else begin
   for i:=0 to 8 do ChannelMap[i].PhysicalChannel:=i;
   for i:=9 to PMusHeader(Buffer)^.Channels-1 do
    ChannelMap[i].PhysicalChannel:=i+1;
  end;
  MusPosition:=PMusHeader(Buffer)^.ScoreStart;
  StartPosition:=MusPosition;
  MusBuffer:=Buffer;
  TicksToWait:=0;
 end
 else MusError:=INVALID_SEQ_FORMAT;
end;


begin
 MusPlaying:=false;
 MusError:=NO_ERROR;
end.

