(************************************************************)
(* OPL2/OPL3 synthesizer low-level interface                *)
(* Coded by Asp //Virtual Research, 1998                    *)
(*                                                          *)
(* Stereo supported for dual-OPL2 (SB Pro I) and for        *)
(* OPL3 (SB Pro II, AdLib Gold, PAS 16 etc.) In both        *)
(* cases maximum of 9 stereo channels available. Of course  *)
(* you also may use 9 mono channels on OPL2 (AdLib, SB).    *)
(* SecondRegSet parameter determines either left or         *)
(* right speaker output for OPL2 stereo and either first or *)
(* second registers set for OPL3 devices. It must be set to *)
(* False for mono devices. Note, that on SB Pro 1.0 (dual   *)
(* OPL2) writing to port $388 causes mono output, while on  *)
(* OPL3 cards port $388 also may be used for stereo         *)
(* output.                                                  *)
(*                                                          *)
(* Unit version : 1.01                                      *)
(* Last updated : 21.Nov.1999                               *)
(************************************************************)

unit OPL;
{$I DEF32.INC}
{$I release.opt}

interface

 const OPL_IO_Addr:word=$388;        (* Must be set to actual synth. addr *)
       OPL_NOT_DETECTED=0;
       OPL_MONO=1;
       OPL2_STEREO=2;
       OPL3_STEREO=3;

 type TInstrument=array[0..10] of byte;

(*
   2-operator instrument definition:
   Offset         Description

   00             Modulator characteristic
   01             Carrier characteristic
            bit 7    Pitch Vibrato
                6    Amplitude Vibrato
                5    Sustaining Sound
                4    Envelope Scaling
                3-0  Frequency Multiplier
   02             Modulator Scaling/Output Level
   03             Carrier Scaling/Output Level
           bit 7-6  Level Scaling
               5-0  Output Level
   04             Modulator Attack/Decay
   05             Carrier Attack/Decay
            bit 7-4 Attack Rate
                3-0 Decay Rate
   06             Modulator Sustain Level/Release Rate
   07             Carrier Sustain Level/Release Rate
            bit 7-4  Sustain Level
                3-0  Release Rate
   08             Modulator Wave Select
   09             Carrier Wave Select
            bit 7-2 (all cleared)
                0-1 Wave Select
   0A             Feedback/Connection
            bit 7-4 (cleared)
                3-1 Modulator Feedback
                0   Connection (AM/FM)
*)

 function  oplDetect:byte;  (* 0 = not detected, 1 = mono, 2 or 3 = stereo *)
 procedure oplReset;
 procedure oplEnableOPL3;  (* Enables OPL3 extensions. Called automatically
               by oplDetect. *)
 procedure oplDisableOPL3;
 function  oplReadStatus(SecondRegSet:boolean):byte;
 procedure oplWriteData(SecondRegSet:boolean;Register,Value:byte);
 procedure oplSetInstrument(SecondRegSet:boolean;Channel:byte;
            Instrument:TInstrument;Left,Right:boolean);
 procedure oplSetInsCharacteristic(SecondRegSet:boolean;
            Channel,Charact,CharactOfs:byte;Left,Right:boolean);
 procedure oplSetVibAMDepth(SecondRegSet:boolean;Vib,AM:boolean);
 procedure oplSetNote(SecondRegSet:boolean;Channel:byte;
            Freq:word;Octave:byte;NoteOn:boolean);

implementation

const RegisterMap:array[0..8,1..2] of byte=( (* Modulator/Carrier *)
       (0,3),(1,4),(2,5),(8,11),(9,12),
       (10,13),(16,19),(17,20),(18,21));

var OldExit:pointer;

procedure oplWriteData(SecondRegSet:boolean;Register,Value:byte);assembler;
asm
 mov dx,OPL_IO_Addr
 cmp SecondRegSet,false
 je @@WriteData
 inc dx
 inc dx
@@WriteData:
 mov al,Register
 out dx,al
 in al,dx
 in al,dx
 in al,dx
 in al,dx
 in al,dx
 in al,dx
 inc dx
 mov al,Value
 out dx,al
 dec dx
{$IFDEF _32BIT}
 mov ecx,35
{$ELSE}
 mov cx,35
{$ENDIF}
@@Delay:
 in al,dx
 loop @@Delay
end;

function oplReadStatus(SecondRegSet:boolean):byte;assembler;
asm
 mov dx,OPL_IO_Addr
 cmp SecondRegSet,false
 je @@ReadStatus
 inc dx
 inc dx
@@ReadStatus:
 in al,dx
end;

procedure oplReset;
var i:byte;
begin
 for i:=2 to $42 do begin
  oplWriteData(false,i,0);
  oplWriteData(true,i,0);
 end;
 for i:=$45 to $55 do begin
  oplWriteData(false,i,$3F);    (* Set volume to zero *)
  oplWriteData(true,i,$3F);
 end;
 for i:=$56 to $F5 do begin
  oplWriteData(false,i,0);
  oplWriteData(true,i,0);
 end;
 oplWriteData(false,1,32); (* Enable waveform select *)
 oplWriteData(true,1,32);
end;

procedure oplEnableOPL3;   (* Enable OPL3 stereo *)
begin
 oplWriteData(true,5,1);
end;

procedure oplDisableOPL3;
begin
 oplWriteData(true,5,0);
end;

function oplDetect:byte;
var i,Result1,Result2:byte;
    OPL3:boolean;
begin
 OPL3:=false;
 oplWriteData(false,4,$60);
 oplWriteData(false,4,$80);
 Result1:=oplReadStatus(false);
 oplWriteData(false,2,$FF);
 oplWriteData(false,4,$21);
 for i:=1 to 150 do Result2:=oplReadStatus(false); (* Read with delay *)
 oplWriteData(false,4,$60);
 oplWriteData(false,4,$80);
 Result1:=Result1 and $E0;
 Result2:=Result2 and $E0;
 if (Result1=0) and (Result2=$C0) then oplDetect:=OPL_MONO
 else begin oplDetect:=OPL_NOT_DETECTED; Exit; end;
   (* Will check for OPL3 *)
 Result1:=oplReadStatus(false);
 if (Result1 and 6)=0 then oplDetect:=OPL3_STEREO;
   (* Now will check for OPL2 stereo *)
 oplWriteData(true,4,$60);
 oplWriteData(true,4,$80);
 Result1:=oplReadStatus(true);
 oplWriteData(true,2,$FF);
 oplWriteData(true,4,$21);
 for i:=1 to 150 do Result2:=oplReadStatus(true);
 oplWriteData(true,4,$60);
 oplWriteData(true,4,$80);
 Result1:=Result1 and $E0;
 Result2:=Result2 and $E0;
 if (Result1=0) and (Result2=$C0) then oplDetect:=OPL2_STEREO;
 oplReset;
 oplEnableOPL3;
end;

procedure oplSetInstrument(SecondRegSet:boolean;Channel:byte;
     Instrument:TInstrument;Left,Right:boolean);
begin
 oplWriteData(SecondRegSet,$20+RegisterMap[Channel,1],Instrument[0]);
 oplWriteData(SecondRegSet,$20+RegisterMap[Channel,2],Instrument[1]);
 oplWriteData(SecondRegSet,$40+RegisterMap[Channel,1],Instrument[2]);
 oplWriteData(SecondRegSet,$40+RegisterMap[Channel,2],Instrument[3]);
 oplWriteData(SecondRegSet,$60+RegisterMap[Channel,1],Instrument[4]);
 oplWriteData(SecondRegSet,$60+RegisterMap[Channel,2],Instrument[5]);
 oplWriteData(SecondRegSet,$80+RegisterMap[Channel,1],Instrument[6]);
 oplWriteData(SecondRegSet,$80+RegisterMap[Channel,2],Instrument[7]);
 oplWriteData(SecondRegSet,$E0+RegisterMap[Channel,1],Instrument[8]);
 oplWriteData(SecondRegSet,$E0+RegisterMap[Channel,2],Instrument[9]);
 if Right then Instrument[10]:=Instrument[10] or 32;
 if Left then Instrument[10]:=Instrument[10] or 16;
 oplWriteData(SecondRegSet,$C0+Channel,Instrument[10]);
end;

procedure oplSetInsCharacteristic(SecondRegSet:boolean;
     Channel,Charact,CharactOfs:byte; Left,Right:boolean);
begin
 case CharactOfs of
  0:oplWriteData(SecondRegSet,$20+RegisterMap[Channel,1],Charact);
  1:oplWriteData(SecondRegSet,$20+RegisterMap[Channel,2],Charact);
  2:oplWriteData(SecondRegSet,$40+RegisterMap[Channel,1],Charact);
  3:oplWriteData(SecondRegSet,$40+RegisterMap[Channel,2],Charact);
  4:oplWriteData(SecondRegSet,$60+RegisterMap[Channel,1],Charact);
  5:oplWriteData(SecondRegSet,$60+RegisterMap[Channel,2],Charact);
  6:oplWriteData(SecondRegSet,$80+RegisterMap[Channel,1],Charact);
  7:oplWriteData(SecondRegSet,$80+RegisterMap[Channel,2],Charact);
  8:oplWriteData(SecondRegSet,$E0+RegisterMap[Channel,1],Charact);
  9:oplWriteData(SecondRegSet,$E0+RegisterMap[Channel,2],Charact);
  10:begin
   if Right then Charact:=Charact or 32;
   if Left then Charact:=Charact or 16;
   oplWriteData(SecondRegSet,$C0+Channel,Charact);
  end;
 end;
end;

procedure oplSetVibAMDepth(SecondRegSet:boolean;Vib,AM:boolean);
var Value:byte;
begin
 Value:=0;
 if Vib then Value:=Value or $80;
 if AM then Value:=Value or $40;
 oplWriteData(SecondRegSet,$BD,Value);
end;

procedure oplSetNote(SecondRegSet:boolean;Channel:byte;
           Freq:word;Octave:byte;NoteOn:boolean);
var Value:byte;
begin
 oplWriteData(SecondRegSet,$A0+Channel,Lo(Freq));
 Value:=((Freq and $0300) shr 8) or ((Octave and 7) shl 2);
 if NoteOn then Value:=Value or 32;
 oplWriteData(SecondRegSet,$B0+Channel,Value);
end;

procedure oplExitProc; {$IFNDEF _32BIT} far; {$ENDIF}
begin
 ExitProc:=OldExit;
 oplReset;
end;

begin
 OldExit:=ExitProc;
 ExitProc:=@oplExitProc;
end.
