(*******************************)
(* User-defined timer ISR      *)
(* (C)oded by Asp //VRg        *)
(* Unit version: 1.56          *)
(* Last updated 17.Sep.1999    *)
(*******************************)

unit Timer;
{$I DEF32.INC}
{$I release.opt}
{$IFNDEF _32BIT}
 {$C FIXED PRELOAD PERMANENT}
{$ENDIF}

interface

procedure TimerSetup(TimesPerSecond:longint);
procedure TimerDone;
procedure TimerReset;
procedure TimerInstallHandler(Handler:pointer);
procedure TimerRemoveHandler;
function  TimerRead:longint;

implementation
uses DOS;

var OldInt8:{$IFDEF __TMT__} farpointer {$ELSE} pointer {$ENDIF};
                        (* pointer to old INT 8 handler *)
    NewInt8:pointer;
    Clock,ClockFlag:word;
    Ticks:longint;

const TimerHandler:pointer=nil;

procedure Int8;interrupt;assembler;
{$IFNDEF _32BIT}
asm
 cmp word ptr TimerHandler,0            (* If user handler <> 0 *)
 jne @@2                                (* then call it *)
 cmp word ptr TimerHandler+2,0          (* else *)
 je @@3                                 (* do not call it *)
@@2:
 push ds                                (* Save data segment *)
 call [TimerHandler]
 pop ds                                 (* Restore data segment *)
@@3:
 mov ax,word ptr Ticks                  (* Inc Ticks *)
 mov bx,word ptr Ticks+2
 add ax,1
 adc bx,0
 mov word ptr Ticks,ax
 mov word ptr Ticks+2,bx
 inc Clock                              (* Inc Clock *)
 mov ax,Clock
 cmp ax,ClockFlag                       (* Is it time for old Int8? *)
 jb @@1
 mov Clock,0                            (* If yes *)
 pushf                                  (* then call old Int8 *)
 call [OldInt8]
 jmp @@exit
@@1:
 mov al,$60                             (* Report end of hardware interrupt *)
 out $20,al
@@exit:
end;
{$ELSE}
asm
 cmp TimerHandler,0                     (* If user handler <> 0 *)
 je @@3                                 (* do not call it *)
 push ds                                (* Save data segment *)
 push es
 call [TimerHandler]
 pop es
 pop ds                                 (* Restore data segment *)
@@3:
 inc Ticks                              (* Inc Ticks *)
 inc Clock                              (* Inc Clock *)
 mov ax,Clock
 cmp ax,ClockFlag                       (* Is it time for old Int8? *)
 jne @@1
 mov Clock,0                            (* If yes *)
 pushfd                                 (* then call old Int8 *)
 call [OldInt8]
 jmp @@exit
@@1:
 mov al,$60                             (* Report end of hardware interrupt *)
 out $20,al
@@exit:
end;
{$ENDIF}

procedure DisableTimerIRQ;assembler;
asm
 in al,$21
 or al,1
 out $21,al
end;

procedure EnableTimerIRQ;assembler;
asm
 in al,$21
 and al,$FE
 out $21,al
end;

procedure TimerSetup(TimesPerSecond:longint);
   (* This one _must_ be called before all *)
var NewCounter:word;
begin
 if TimesPerSecond<19 then TimesPerSecond:=19;
 if TimesPerSecond>1193180 then TimesPerSecond:=1193180;
 NewCounter:=1193180 div TimesPerSecond;
 ClockFlag:=TimesPerSecond*1000 div 18206;
 NewInt8:=@Int8;
 Ticks:=0;
 Clock:=0;
 DisableTimerIRQ;
 asm
  mov al,$36      (* Set new time counter *)
  out $43,al
  mov bx,NewCounter
  mov al,bl
  out $40,al
  mov al,bh
  out $40,al
 end;
 SetIntVec(8,NewInt8);
 EnableTimerIRQ;
end;

procedure TimerDone;
   (* This one MUST be called after all *)
begin
 DisableTimerIRQ;
 asm
  mov al,$36      (* Set old time counter *)
  out $43,al
  xor ax,ax
  out $40,al
  out $40,al
 end;
 SetIntVec(8,OldInt8);
 EnableTimerIRQ;
end;

procedure TimerReset;
   (* Reset Ticks to zero *)
begin
 Ticks:=0;
end;

function TimerRead:longint;
   (* Read number of timer ticks passed
      from last timerReset *)
begin
 TimerRead:=Ticks;
end;

procedure TimerInstallHandler(Handler:pointer);
   (* Handler must point to procedure with no input parameters *)
   (* Theoretically you even can swap handlers after TimerSetup,
      though this is not recommended *)
begin
 DisableTimerIRQ;
 TimerHandler:=Handler;
 EnableTimerIRQ;
end;

procedure TimerRemoveHandler;
begin
 DisableTimerIRQ;
 TimerHandler:=nil;
 EnableTimerIRQ;
end;

begin
 GetIntVec(8,OldInt8);
end.
