(*******************************************)
(* VRMMP MIDI Player v. 1.12               *)
(* Copyleft (C) 1998-1999 Virtual Research *)
(* Finished 03.Dec.1998                    *)
(* Last updated 23.Nov.1999                *)
(*******************************************)

program VRMMP;

{$I def32.inc}
{$IFNDEF _32BIT}
 {$F+}
 {$C FIXED PRELOAD PERMANENT}
{$ENDIF}
{$I release.opt}

uses VRMMSDef,MUS,MID,PStrings,CRT,DOS;

{$IFDEF _32BIT} {$DEFINE _PM} {$ENDIF}
{$IFDEF DPMI}   {$DEFINE _PM} {$ENDIF}

{$IFNDEF _PM}
{$I DosShel.inc}
{$ENDIF}

const Version='1.12';
      SB_PORTS=8;
      SBPorts:array[1..8] of word=($220,$240,$210,$230,$250,$260,$280,$388);
      HELP_STRINGS=17;
      HelpScreen:array[1..HELP_STRINGS] of pchar=(
       'Usage: VRMMP <midi file> [options]',
       '',
       '<midi file> is MUS or MID file name;',
       ' options are:',
       '/d<device>',
       ' where <device> is a number, one of the following:',
       ' 0 = Debug (MIDI stream dumper)',
       ' 1 = OPL2 FM synthesizer',
       ' 2 = OPL3 FM synthesizer',
       ' 3 = Roland MPU401 or compatible;',
       '/p<port>',
       ' where <port> is sound device port (hexadecimal);',
       '/t<timer rate>',
       ' where <timer rate> is a number in range [1..10];',
       '/m<mus tempo>',
       ' where <mus tempo> is a MUS sequence playback speed, TPS;',
       '/l - if specified, loop music for ever.');
       AUTODETECT=-1;
       TITLE=NL+'VRMMP MIDI player, v. '+Version+NL+
                'Copyleft (C) 1998-1999 Virtual Research'+NL;

       type TSeqType=(STMus, STMid);


var Sequence:pointer;
    MusSize:word;
    MidiPort:word;
    Device:integer;
    Vol:byte;
    FileName:string;
    Dummy:integer;
    Loop:boolean;
    SeqType:TSeqType;
    C:char;


procedure ShowHelp;
var i:word;
begin
 for i:=1 to HELP_STRINGS do WriteLn(HelpScreen[i]);
 Halt;
end;

procedure AnalyzeCmdLine;
var Param:string;
    Code:integer;
    i:word;
begin
 if ParamCount=0 then ShowHelp;
 FileName:=''; Loop:=false;
 Device:=AUTODETECT; MidiPort:=0;
 for i:=1 to ParamCount do begin
  Param:=ParamStr(i);
  if (Param[1]<>'/') and (Param[1]<>'-') then begin
   if FileName='' then FileName:=UpCaseStr(Param);
  end else
   case UpCase(Param[2]) of
    'D':begin
     if Length(Param)>3 then ShowHelp;
     case Param[3] of
      '0':Device:=_Debug;
      '1':Device:=_AdLib;
      '2':Device:=_SBPro;
      '3':Device:=_MPU401;
     else ShowHelp;
     end;
    end;
    'P':begin
     Delete(Param,1,2);
     Param:='$'+Param;
     Val(Param,MidiPort,Code);
     if Code<>0 then ShowHelp;
    end;
    'T':begin
     Delete(Param,1,2);
     Val(Param,Dummy,Code);
     if (Code<>0) or
        (Dummy<1) or
        (Dummy>10) then ShowHelp;
     MusSetTimerRate(182*Dummy);
     MidSetTimerRate(182*Dummy);
    end;
    'M':begin
     Delete(Param,1,2);
     Val(Param,Dummy,Code);
     if (Code<>0) then ShowHelp;
     MusSetTempo(Dummy);
    end;
    'L':Loop:=true;
   else ShowHelp;
   end;
 end;
end;

type TInitDevice=procedure(Device,Port:word);
     TShutDownDevice=procedure;

var InitDevice:TInitDevice;
    ShutDownDevice:TShutDownDevice;
    Error:integer;

procedure SetupProcs;
begin
 if SeqType=STMus then begin
  InitDevice:=MusInitDevice;
  ShutDownDevice:=MusShutDownDevice;
 end
 else begin
  InitDevice:=MidInitDevice;
  ShutDownDevice:=MidShutDownDevice;
 end;
end;

procedure SetError;
begin
 if SeqType=STMus then Error:=MusError else Error:=MidError;
end;

function DetectGM:boolean;
begin
 DetectGM:=false;
 InitDevice(_MPU401,$330);
 SetError; if Error=0 then begin DetectGM:=true; Exit; end;
 InitDevice(_MPU401,$300);
 SetError; if MusError=0 then DetectGM:=true;
end;

function DetectOPL3:boolean;
var i:byte;
begin
 DetectOPL3:=false;
 for i:=1 to SB_PORTS do begin
  InitDevice(_SBPro,SBPorts[i]);
  SetError; if Error=0 then begin DetectOPL3:=true; break; end;
 end;
end;

function DetectOPL2:boolean;
var i:byte;
begin
 DetectOPL2:=false;
 for i:=1 to 8 do begin
  InitDevice(_SB,SBPorts[i]);
  SetError; if Error=0 then begin DetectOPL2:=true; break; end;
 end;
end;

procedure CheckError;
begin
 SetError;
 if Error<>NO_ERROR then begin
  WriteLn(ErrorMsg[Error]);
  Halt(255);
 end;
end;

procedure LoadSequence;
begin
 (* First try to load as MUS *)
 MusLoad(FileName,0,Sequence,MusSize);
 if MusError=NO_ERROR then begin
  SeqType:=STMus;
  MusInitSequence(Sequence);
  CheckError;
 end
 else begin
  (* Now try to load as MID *)
  SeqType:=STMid;
  MidLoad(FileName,0,Sequence);
  CheckError;
  MidInitSequence(Sequence);
  CheckError;
 end;
end;

procedure PlaySequence;
begin
 if SeqType=STMus then MusPlay(Loop) else MidPlay(Loop);
end;

procedure ShutDown;
begin
 if SeqType=STMus then begin
  if Device<>_Debug then MusStop;
  FreeMem(Sequence,MusSize);
 end
 else begin
  if Device<>_Debug then MidStop;
  MidFree(Sequence);
 end;
 ShutDownDevice;
 ClrScr;
end;

procedure SetVolume(NewVol:byte);
begin
 if SeqType=STMus then MusSetVolume(NewVol) else MidSetVolume(NewVol);
end;

procedure PlayerScreen;
begin
  ClrScr;
  WriteLn(TITLE);
  WriteLn('Playing MIDI sequence...',NL,
	  'Ŀ',NL,
	  ' <esc> = stop            ',NL,
	  ' <d>   = DOS shell       ',NL,
	  ' <'#27'>   = decrease volume ',NL,
	  ' <'#26'>   = increase volume ',NL,
	  '' );
  WriteLn('Volume: ',StrStr('',Vol),StrStr(' ',64-Vol),'');
end;

begin
 WriteLn(TITLE);
 AnalyzeCmdLine;
 Writeln('Loading MIDI sequence...');
 LoadSequence;
 Writeln('Initializing MIDI device...');
 SetupProcs;
 if Device<>AUTODETECT then begin
  if MidiPort=0 then begin
   case Device of
    _Debug: InitDevice(Device,0);
    _AdLib: DetectOPL2;
    _SBPro: DetectOPL3;
    _MPU401:DetectGM;
   end;
  end
  else InitDevice(Device,MidiPort);
  CheckError;
 end
 else begin
  if not DetectGM then
   if not DetectOPL3 then DetectOPL2;
  CheckError;
 end;
 while KeyPressed do ReadKey;   (* Clear keyboard buffer *)
 if Device<>_Debug then begin
  Vol:=64; PlayerScreen;
  PlaySequence;
  while C<>#27 do begin
   if KeyPressed then C:=ReadKey else C:=#$FF;
   case C of
    #0:begin
     C:=ReadKey;
     case C of
      #75:begin          (* Left arrow *)
       if vol>0 then begin
        Dec(Vol); GotoXY(10,WhereY-1);
        WriteLn(StrStr('',vol)+StrStr(' ',64-vol),'');
       end;
       SetVolume(Vol);
      end;
      #77:begin           (* Right arrow *)
       if Vol<64 then begin
        Inc(Vol); GotoXY(10,WhereY-1);
        WriteLn(StrStr('',Vol)+StrStr(' ',64-Vol),'');
       end;
       SetVolume(vol);
      end;
     end;
    end;
    'd','D':begin
     ClrScr;
{$IFNDEF _PM}
     if DosShell('')<>0 then begin
{$ELSE}
     DosError:=0;       (* For TMT 2.x *)
     SwapVectors;
     Exec(GetEnv('COMSPEC'),'');
     SwapVectors;
     if DosError<>0 then begin
{$ENDIF}
      WriteLn('Error executing command interpreter!',NL,'Press a key...');
      ReadKey;
     end;
     PlayerScreen;
    end;
   end;
   if ((SeqType=STMus) and (not MusPlaying)) or
      ((SeqType=STMid) and (not MidPlaying)) then Break;
  end;
 end
 else begin
  Assign(Output,''); Rewrite(Output);
  if SeqType=STMus then begin
   MusPlaying:=true;
   MusLoop:=Loop;
  end
  else begin
   MidPlaying:=true;
   MidLoop:=Loop;
  end;
  while C<>#27 do begin
   if SeqType=STMus then MusPoll else MidPoll;
   if KeyPressed then C:=ReadKey;
   if ((SeqType=STMus) and (not MusPlaying)) or
      ((SeqType=STMid) and (not MidPlaying)) then Break;
  end;
 end;
 ShutDown;
end.
